/*
 *   (C) Copyright IBM Corp. 2001, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: md_create.c
 *
 * Description: This file contains all functions related to the common creation
 *              of MD volumes.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <plugin.h>

#include "md.h"

#define my_plugin_record my_plugin

static u_int32_t get_random(void) {
	u_int32_t num;
	int fd;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd != -1 && read(fd, &num, 4) == 4) {
		LOG_DEBUG("raid set magic: %x\n", num);
		close(fd);
	} else {
		num = rand();
		LOG_DEBUG("raid set magic (pseudo-random): %x\n", num);
	}
	close(fd);
	return num;
}

int md_clone_superblock( md_volume_t *volume,int index) {
	int rc = 0;
	LOG_ENTRY();
	if (!(volume->super_array[index] =  EngFncs->engine_alloc(MD_SB_BYTES))) {
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
	*(volume->super_array[index]) = *(volume->super_block);
	volume->super_array[index]->this_disk = volume->super_array[index]->disks[index];

	LOG_EXIT_INT(rc);
	return rc;
}





/*
 md_create_first_superblock :
		Allocate and initialize a super block, filling in all of the common fields.
	Caller is responsible for filling in the following before cloning SB:


*/


int md_create_first_superblock( md_volume_t *volume,
				mdp_disk_t first_disk,
				u_int32_t level,
				u_int32_t chunk_size,
				u_int32_t size,
				u_int32_t nr_disks,
				u_int32_t spare_disks,
				u_int32_t state) {

	int rc = 0;
	int i,j;
	mdp_super_t * super;
	md_volume_t * tmp_volume = volume_list_head;
	char minor_array[MAX_MD_MINORS];
	LOG_ENTRY();

	if (!(super = EngFncs->engine_alloc(MD_SB_BYTES ))) {
		LOG_CRITICAL("Memory error creating buffer for new super block.\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	memset(minor_array, 0x0, sizeof(minor_array));
	/*
	 * For each of the MD volumes, indicate the corresponding MD minor has been taken.
	 */
	while (tmp_volume != NULL) {
		minor_array[tmp_volume->super_block->md_minor] = 1;
		tmp_volume = tmp_volume->next;
	}

	/*
	 * Walk the minor_array[], find an available slot and 
	 * query the kernel to make sure that the MD minor is really available.
	 */
	for (i=0; i < MAX_MD_MINORS; i++) {
		storage_object_t *region;
		int rc2;
		char region_name[7];
		mdu_array_info_t info;

		if (minor_array[i] == 0) {
			sprintf(region_name, "md/md%d", i);
			rc2 = EngFncs->allocate_region(region_name, &region);
			if (rc2 == 0) {
				/*
				 * Query the kernel MD driver.
				 * If the md minor is not in use, we should get an error.
				 */
				region->dev_major = MD_MAJOR;
				region->dev_minor = i;
				rc2 = md_ioctl_get_array_info(region, &info);
				EngFncs->free_region(region);
				if (rc2) {
					break;
				}
			}
		}
	}

	if (i >= MAX_MD_MINORS) {
		LOG_ERROR("No more MD devices avaliable, creation failing");
		EngFncs->engine_free(super);
		LOG_EXIT_INT(ENODEV);
		return ENODEV;
	}

	volume->super_block = super;

	//cluster naming support
	sprintf(volume->name, "md/md%d",i);
	for (j=0; j<nr_disks; j++){
		if ( volume->child_object[j] != NULL ) {
			if (volume->child_object[j]->disk_group) {
				sprintf(volume->name, 
					"%s/md/md%d",
					volume->child_object[j]->disk_group->name,i);
			}
		}
	}
	
	super->active_disks = nr_disks - spare_disks;
	super->chunk_size = chunk_size * BLOCK_SIZE;
	super->ctime = super->utime =  time(NULL);
	for (j = 0; j < nr_disks; j++) {
		first_disk.major = volume->child_object[j]->dev_major;
		first_disk.minor = volume->child_object[j]->dev_minor;
		first_disk.number = j;
		first_disk.raid_disk = j;
		super->disks[j] = first_disk;
	}
	super->events_lo = 1;
	super->events_hi = 0;
	super->failed_disks = 0;
//	super->gstate_creserved -- Padding space, not used
//	super->gstate_sreserved -- Padding space, not used
//	super->gvalid_words     -- does not appear to ever be filled in
	super->layout = 0;	   // parity layout for raid 5 see *parity_algorithm_table[] in parser.c
	super->level = level;
	super->major_version = 0;
	super->md_magic = MD_SB_MAGIC;
	super->md_minor = i;
	super->minor_version = 90;
	super->not_persistent = 0;
	super->nr_disks = nr_disks;
	super->patch_version = 0;
//	super->pstate_reserved-- Padding space, not used
	super->raid_disks = nr_disks - spare_disks;
//	super->reserved = -- Padding space, not used
//	super->root_block -- Does not appear to be set
//	super->root_pv -- Does not appear to be set
	super->sb_csum = 0;	  // Set on commit
	super->set_uuid0 = get_random();
	super->set_uuid1 = get_random();
	super->set_uuid2 = get_random();
	super->set_uuid3 = get_random();
	super->size = size;
	super->spare_disks = spare_disks;
	super->state = state;
	super->this_disk = super->disks[0];
//	super->utime -- filled in above
	super->working_disks = nr_disks;

	LOG_EXIT_INT(rc);
	return rc;
}



